<?php

namespace Drupal\Tests\menu_link\Kernel;

use Drupal\Core\Database\Database;
use Drupal\Core\DrupalKernel;
use Drupal\Core\Extension\ExtensionDiscovery;
use Drupal\Core\Menu\MenuTreeParameters;
use Drupal\Core\Site\Settings;
use Drupal\entity_test\Entity\EntityTestMul;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\KernelTests\KernelTestBase;
use PHPUnit\Framework\Exception;
use Drupal\TestTools\Comparator\MarkupInterfaceComparator;
use Symfony\Cmf\Component\Routing\RouteObjectInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Routing\Route;

/**
 * Tests the menu link field formatters.
 *
 * @group menu_link
 */
class MenuLinkFormatterTest extends KernelTestBase {

  /**
   * {@inheritdoc}
   */
  protected static $modules = ['menu_link', 'entity_test', 'field', 'user', 'system', 'language'];

  /**
   * {@inheritdoc}
   */
  protected function setUp() : void {
    parent::setUp();

    // Allow tests to compare MarkupInterface objects via assertEquals().
    $this->registerComparator(new MarkupInterfaceComparator());

    $this->root = static::getDrupalRoot();
    $this->initFileCache();
    $this->bootEnvironment();
    $this->bootKernel();
  }

  /**
   * Installs database tables from a module schema definition.
   *
   * @param string $module
   *   The name of the module that defines the table's schema.
   * @param string|array $tables
   *   The name or an array of the names of the tables to install.
   *
   * @throws \LogicException
   *   If $module is not enabled or the table schema cannot be found.
   */
  protected function installSchema($module, $tables) : void {
    $this->installEntitySchema('user');
    $this->installEntitySchema('entity_test_mul');

    FieldStorageConfig::create([
      'field_name' => 'field_menu_link',
      'entity_type' => 'entity_test_mul',
      'type' => 'menu_link',
    ])->save();

    FieldConfig::create([
      'field_name' => 'field_menu_link',
      'entity_type' => 'entity_test_mul',
      'bundle' => 'entity_test_mul',
    ])->save();

  }

  /**
   * Tests the menu_link and menu_link_breadcrumb field formatters.
   */
  public function testMenuLinkFormatters() : void {
    $entity_test_mul = EntityTestMul::create([
      'type' => 'entity_test_mul',
      'name' => 'test',
      'field_menu_link' => [
        'menu_name' => 'test_menu',
        'title' => 'test title 1',
        'description' => 'test description',
      ],
    ]);

    /** @var \Drupal\Core\Menu\MenuLinkTree $menu_tree */
    $menu_tree = \Drupal::service('menu.link_tree');
    $parameters = new MenuTreeParameters();
    $parameters->addCondition('title', 'test title 1');
    $entity_test_mul->save();
    $result = $menu_tree->load('test_menu', $parameters);
    $menu_link = reset($result);
    // Add another entity as a child of the first one.
    $entity_test_mul2 = EntityTestMul::create([
      'type' => 'entity_test_mul',
      'name' => 'test',
      'field_menu_link' => [
        'menu_name' => 'test_menu',
        'parent' => $menu_link->link->getPluginId(),
        'title' => 'test title 2',
        'description' => 'test description 2',
      ],
    ]);
    $entity_test_mul2->save();

    // Test that the menu_link formatter outputs plain text.
    $content = $this->renderLink($entity_test_mul2, 'menu_link', ['link_to_target' => FALSE]);
    $this->assertRaw('<div>test title 2</div>', $content);

    // Test that the menu_link formatter outputs a link.
    $content2 = $this->renderLink($entity_test_mul2, 'menu_link', ['link_to_target' => TRUE]);
    $this->assertRaw('<div><a href="/entity_test_mul/manage/2" title="test description 2" hreflang="en">test title 2</a></div>', $content2);

    // Test that the breadcrumb formatter outputs links.
    $content3 = $this->renderLink($entity_test_mul2, 'menu_link_breadcrumb', ['link_to_target' => TRUE]);
    $this->assertRaw('<a href="/entity_test_mul/manage/1">test title 1</a>', $content3);
    $this->assertRaw('<a href="/entity_test_mul/manage/2">test title 2</a>', $content3);

    // Test that the breadcrumb formatter outputs plain text.
    $content4 = $this->renderLink($entity_test_mul2, 'menu_link_breadcrumb', ['link_to_target' => FALSE]);
    $this->assertRaw('test title 1', $content4);
    $this->assertRaw('test title 2', $content4);
    $this->assertNoRaw('<a href="/entity_test_mul/manage/1">test title 1</a>', $content3);
    $this->assertNoRaw('<a href="/entity_test_mul/manage/2">test title 2</a>', $content3);

    // Test that we skip the last element when parents_only is enabled.
    $content5 = $this->renderLink($entity_test_mul2, 'menu_link_breadcrumb', ['link_to_target' => FALSE, 'parents_only' => TRUE]);
    $this->assertRaw('test title 1', $content5);
    $this->assertNoRaw('test title 2', $content5);
  }

  /**
   * Renders a menu link field.
   *
   * @param $entity
   *   The entity to render the link for.
   * @param string|array $type
   *   The ID of the field formatter.
   * @param string|array $settings
   *   The field formatter settings.
   *
   * @return string
   *   The rendered output.
   */
  protected function renderLink($entity, $type, $settings) : string {
    $display = \Drupal::service('entity_display.repository')->getViewDisplay('entity_test_mul', 'entity_test_mul', 'default')
      ->setComponent('field_menu_link', [
        'type' => $type,
        'settings' => $settings,
      ]);
    $content = $display->build($entity);
    return $this->render($content);
  }

  /**
   * Returns the modules to enable for this test.
   *
   * @param string $class
   *   The fully-qualified class name of this test.
   *
   * @return array
   */
  private static function getModulesToEnable(string $class) : array {
    $modules = [];
    while ($class) {
      if (property_exists($class, 'modules')) {
        // Only add the modules, if the $modules property was not inherited.
        $rp = new \ReflectionProperty($class, 'modules');
        if ($rp->class === $class) {
          $modules[$class] = $class::$modules;
        }
      }
      $class = get_parent_class($class);
    }
    // Modules have been collected in reverse class hierarchy order; modules
    // defined by base classes should be sorted first. Then, merge the results
    // together.
    $modules = array_reverse($modules);
    return call_user_func_array('array_merge_recursive', $modules);
  }

  /**
   * Returns Extension objects for $modules to enable.
   *
   * @param string[] $modules
   *   The list of modules to enable.
   *
   * @return \Drupal\Core\Extension\Extension[]
   *   Extension objects for $modules, keyed by module name.
   *
   * @throws \PHPUnit\Framework\Exception
   *   If a module is not available.
   *
   * @see \Drupal\Tests\KernelTestBase::enableModules()
   * @see \Drupal\Core\Extension\ModuleHandler::add()
   */
  private function getExtensionsForModules(array $modules) : array {
    $extensions = [];
    $discovery = new ExtensionDiscovery($this->root);
    $discovery->setProfileDirectories([]);
    $list = $discovery->scan('module');
    foreach ($modules as $name) {
      if (!isset($list[$name])) {
        throw new Exception("Unavailable module: '$name'. If this module needs to be downloaded separately, annotate the test class with '@requires module $name'.");
      }
      $extensions[$name] = $list[$name];
    }
    return $extensions;
  }

  /**
   * Bootstraps a kernel for a test.
   */
  private function bootKernel() : void {
    $this->setSetting('container_yamls', []);
    // Allow for test-specific overrides.
    $settings_services_file = $this->root . '/sites/default/testing.services.yml';
    if (file_exists($settings_services_file)) {
      // Copy the testing-specific service overrides in place.
      $testing_services_file = $this->siteDirectory . '/services.yml';
      copy($settings_services_file, $testing_services_file);
      $this->setSetting('container_yamls', [$testing_services_file]);
    }

    // Allow for global test environment overrides.
    if (file_exists($test_env = $this->root . '/sites/default/testing.services.yml')) {
      $GLOBALS['conf']['container_yamls']['testing'] = $test_env;
    }
    // Add this test class as a service provider.
    $GLOBALS['conf']['container_service_providers']['test'] = $this;

    $modules = self::getModulesToEnable(get_class($this));

    // Bootstrap the kernel. Do not use createFromRequest() to retain Settings.
    $kernel = new DrupalKernel('testing', $this->classLoader, FALSE);
    $kernel->setSitePath($this->siteDirectory);
    // Boot a new one-time container from scratch. Set the module list upfront
    // to avoid a subsequent rebuild or setting the kernel into the
    // pre-installer mode.
    $extensions = $modules ? $this->getExtensionsForModules($modules) : [];
    $kernel->updateModules($extensions, $extensions);

    // DrupalKernel::boot() is not sufficient as it does not invoke preHandle(),
    // which is required to initialize legacy global variables.
    $request = Request::create('/');
    $kernel->boot();
    $request->attributes->set(RouteObjectInterface::ROUTE_OBJECT, new Route('<none>'));
    $request->attributes->set(RouteObjectInterface::ROUTE_NAME, '<none>');
    $kernel->preHandle($request);

    $this->container = $kernel->getContainer();

    // Ensure database tasks have been run.
    require_once __DIR__ . '/../../../../../../core/includes/install.inc';
    $connection_info = Database::getConnectionInfo();
    $driver = $connection_info['default']['driver'];
    $namespace = $connection_info['default']['namespace'] ?? NULL;
    $errors = db_installer_object($driver, $namespace)->runTasks();
    if (!empty($errors)) {
      self::fail('Failed to run installer database tasks: ' . implode(', ', $errors));
    }

    if ($modules) {
      $this->container->get('module_handler')->loadAll();
    }

    // Setup the destination to the be frontpage by default.
    \Drupal::destination()->set('/');

    // Write the core.extension configuration.
    // Required for ConfigInstaller::installDefaultConfig() to work.
    $this->container->get('config.storage')->write('core.extension', [
      'module' => array_fill_keys($modules, 0),
      'theme' => [],
      'profile' => '',
    ]);

    $settings = Settings::getAll();
    $settings['php_storage']['default'] = [
      'class' => '\Drupal\Component\PhpStorage\FileStorage::class',
    ];
    new Settings($settings);

    // Manually configure the test mail collector implementation to prevent
    // tests from sending out emails and collect them in state instead.
    // While this should be enforced via settings.php prior to installation,
    // some tests expect to be able to test mail system implementations.
    $GLOBALS['config']['system.mail']['interface']['default'] = 'test_mail_collector';

    // Manually configure the default file scheme so that modules that use file
    // functions don't have to install system and its configuration.
    // @see file_default_scheme()
    $GLOBALS['config']['system.file']['default_scheme'] = 'public';
  }

}
